//
//  NSObject+Reflection.m
//  ReflectionHelper
//
//  Created by Andreas Gerlach on 9/29/12.
//  Copyright (c) 2012 Appelgriebsch. All rights reserved.
//

#import <objc/runtime.h>
#import "NSObject+Reflection.h"
#import "NSData+Base64.h"
#import "NSObject+UniqueIdentifiable.h"

@implementation NSObject (Reflection)

/*!
    @return the dictionary with the name / value of the properties of an object
 */
- (NSDictionary *)properties {

    const char *szClassName = class_getName([self class]);
    NSString *className = [NSString stringWithCString:szClassName encoding:NSASCIIStringEncoding];

    int classVersion = class_getVersion([self class]);

    NSMutableDictionary *dct = [self inspectObject:self ofClass:[self class]];

    if (dct) {

        if ([dct count] == 0) [dct setObject:[self description] forKey:DICT_KEY_DESCRIPTION];

        [dct setObject:[self uniqueObjectID] forKey:DICT_KEY_ID];

        [dct setObject:className forKey:DICT_KEY_NAME];
        [dct setObject:[NSNumber numberWithInt:classVersion] forKey:DICT_KEY_VERSION];

        if (([self respondsToSelector:@selector(encodeWithCoder:)]) &&
                (![[self class] isSubclassOfClass:[NSData class]])) {

            NSData *data = [NSKeyedArchiver archivedDataWithRootObject:self];

            if (data) {

                NSString *base64String = [data base64Encoding];
                [dct setObject:base64String forKey:DICT_KEY_DATA];
            }
        }
    }

    return dct;
}

#pragma mark Internal Helpers

/*!
 checks the properties of the given object and
 builds a dictionary containing property name and value
 
 it will include all properties from the base class - going up the hierarchy until it examines the base class NSObject
 
 @author andreas.gerlach
 
 @param theObject - the object to examine
 @param ofClass - the class definition used as base for this object
 
 @return a dictionary containing property-name / value pairs
 
 */
- (NSMutableDictionary *)inspectObject:(id)theObject
                               ofClass:(Class)theClass {

    NSMutableDictionary *propDict = [[NSMutableDictionary alloc] init];

    unsigned int cnt;
    objc_property_t *propList = class_copyPropertyList(theClass, &cnt);

    for (int i = 0; i < cnt; ++i) {

        objc_property_t theProp = propList[i];

        const char *szPropName = property_getName(theProp);

        if (szPropName) {

            NSString *propertyName = [NSString stringWithCString:szPropName encoding:NSASCIIStringEncoding];

            @try {

                id value = [theObject valueForKey:propertyName];

                if ([[value class] isSubclassOfClass:[NSSet class]]) {

                    [propDict setObject:[value allObjects] forKey:propertyName];            // handle NSSet related stuff as simple array
                }
                else {

                    [propDict setObject:value forKey:propertyName];
                }
            }
            @catch (NSException *exception) {

                NSLog(@"ERROR reflecting class %@: %@", theClass, [exception description]);
                break;
            }
        }

    }

    free(propList);

    //recursively call into the super class
    Class superClass = class_getSuperclass(theClass);
    if ((superClass != nil) && ![superClass isEqual:[NSObject class]]) {

        NSMutableDictionary *superProps = [self inspectObject:theObject ofClass:superClass];
        [propDict setValuesForKeysWithDictionary:superProps];
    }

    return propDict;
}

@end
