//
//  mongo_objc.m
//  mongo-objc
//
//  Created by Andreas Gerlach on 10/25/12.
//  Copyright (c) 2012 Appelgriebsch. All rights reserved.
//

#import "mongo_objc.h"
#import "mongo.h"
#import <ReflectionHelper/ReflectionHelper.h>

@interface MongoDB()

@property(nonatomic, readonly) mongo* mongodb;

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval;

@end


@implementation MongoDB

#pragma mark Deinitialization

- (void)cleanup {
    
    if (_mongodb) {
        
        mongo_destroy(_mongodb);
        free(_mongodb);
        
        _mongodb = NULL;
    }
}

- (void)dealloc {
    
    [self cleanup];
}

- (void)disconnect {
    
    [self cleanup];
}

#pragma mark Initialization

+ (id)connectTo:(NSString *)host onPort:(NSUInteger)port {
 
    MongoDB *mongo = [[MongoDB alloc] initWithHost:host onPort:port withTimeout:-1];
    return mongo;
}

+ (id)connectTo:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval {
    
    MongoDB *mongo = [[MongoDB alloc] initWithHost:host onPort:port withTimeout:interval];
    return mongo;
}

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval {
    
    self = [super init];
    
    if (self) {
        
        _mongodb = malloc(sizeof(mongo));
        
        mongo_init(_mongodb);
        
        if (interval > 0) {
            
            mongo_set_op_timeout(_mongodb, interval);
        }
        
        mongo_connect(_mongodb, [host cStringUsingEncoding:NSUTF8StringEncoding], (int)port);
    }
    
    return self;
}

#pragma mark Properties

- (BOOL)isConnected {
    
    if (self.mongodb == NULL)
        return FALSE;
    
    return (self.mongodb->connected);
}


#pragma mark Methods

- (NSString *)storeDocument:(id)document intoCollection:(NSString *)collection {
    
    NSString *result = nil;
    
    bson *doc = [self buildBSONDocument:document];
    
    if (doc) {
        
        int insert = mongo_insert(self.mongodb, [collection cStringUsingEncoding:NSUTF8StringEncoding], doc, NULL);
        
        bson_destroy(doc);
        
        free(doc);
        
        result = (insert == MONGO_OK ? [document uniqueObjectID] : nil);
    }
    
    return result;
}

- (id)retrieveObjectForDocument:(NSString *)docID fromCollection:(NSString *)collection {
    
    bson doc;
    bson *query = [self buildBSONQuery:docID];
    
    id result = nil;
    
    if (query) {

        int found = mongo_find_one(self.mongodb, [collection cStringUsingEncoding:NSUTF8StringEncoding], query, NULL, &doc);
        
        bson_destroy(query);
        free(query);
        
        if (found == MONGO_OK) {
            
            NSDictionary *dictProperties = [self buildObjectFromBSON:&doc];
            
            if (dictProperties) {
                
                result = [ReflectionHelper initObjectWithValues:dictProperties];
            }
        }
    }
    
    return result;
}

- (bson *) buildBSONDocument:(id)document {
    
    bson* doc = NULL;
    NSDictionary *dct = [ReflectionHelper inspectObject:document];
    
    if (dct) {

        doc = malloc(sizeof(bson));
        
        bson_init(doc);
        
        for (NSString* key in [dct allKeys]) {
            
            id value = [dct objectForKey:key];
            
            if ([[value class] isSubclassOfClass:[NSString class]]) {
                
                bson_append_string(doc, [key cStringUsingEncoding:NSUTF8StringEncoding], [value cStringUsingEncoding:NSUTF8StringEncoding]);
            }
            else if ([[value class] isSubclassOfClass:[NSNumber class]]) {
                
                bson_append_double(doc, [key cStringUsingEncoding:NSUTF8StringEncoding], [value doubleValue]);
            }
            else if ([[value class] isSubclassOfClass:[NSDate class]]) {
                
                bson_append_date(doc, [key cStringUsingEncoding:NSUTF8StringEncoding], [value timeIntervalSince1970]);
            }
            else if ([[value class] isSubclassOfClass:[NSArray class]]) {
                
                bson_append_start_array(doc, [key cStringUsingEncoding:NSUTF8StringEncoding]);
                
                for (id elem in value) {
                    
                    //TODO:
                }
                
                bson_append_finish_array(doc);
            }
            else {
                
                bson* sub = [self buildBSONDocument:value];
                bson_append_bson(doc, [key cStringUsingEncoding:NSUTF8StringEncoding], sub);
            }
        }
        
        bson_finish(doc);
    }
    
    return doc;
}

-(bson *)buildBSONQuery:(NSString *)documentID {
    
    bson *query = malloc(sizeof(bson));
    bson_init(query);
    
    bson_append_string(query, "__id__", [documentID cStringUsingEncoding:NSUTF8StringEncoding]);
    
    bson_finish(query);
    return query;
}

-(id)buildObjectFromBSON:(bson *)doc {
    
    id result = [NSMutableDictionary dictionary];
    
    bson_iterator iter;
    bson_iterator_init(&iter, doc);
    
    while(bson_iterator_more(&iter)) {
        
        bson_type t = bson_iterator_next(&iter);
        const char* key = bson_iterator_key(&iter);
        
        switch(t) {
                
            case BSON_ARRAY:
                break;
                
            case BSON_BINDATA:
                break;
                
            case BSON_BOOL:
                
                [result setValue:[NSNumber numberWithBool:bson_iterator_bool(&iter)]
                          forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                break;
                
            case BSON_DATE:
                
                [result setValue:[NSDate dateWithTimeIntervalSince1970:bson_iterator_date(&iter)]
                          forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                break;
                
            case BSON_DOUBLE:
                
                [result setValue:[NSNumber numberWithDouble:bson_iterator_double(&iter)]
                          forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                break;
                
            case BSON_INT:
                
                [result setValue:[NSNumber numberWithInt:bson_iterator_int(&iter)]
                          forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                break;
                
            case BSON_LONG:
                
                [result setValue:[NSNumber numberWithLong:bson_iterator_long(&iter)]
                          forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                break;
                
            case BSON_OBJECT:
                
                if (true) {
                    
                    bson subObj;
                    bson_iterator_subobject(&iter, &subObj);
                    [result setValue:[self buildObjectFromBSON:&subObj]
                              forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                }
                break;
                
            case BSON_STRING:
                
                if (true) {

                    const char* value = bson_iterator_string(&iter);
                
                    [result setValue:[NSString stringWithCString:value encoding:NSUTF8StringEncoding]
                              forKey:[NSString stringWithCString:key encoding:NSUTF8StringEncoding]];
                }
                break;
                
            default:
                break;
        }
    }
    
    return result;
}
@end
