//
//  redis-objc.m
//  redis-sample
//
//  Created by Andreas Gerlach on 7/8/12.
//  Copyright (c) 2012 Appelgriebsch. All rights reserved.
//

#import "redis-objc.h"
#import "hiredis.h"

@interface APGRedisServer ()

@property(nonatomic, readonly) redisContext *context;

- (id)initWithUnixPath:(NSString *)unixPath;

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port;

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval;

- (redisReply *)execute:(NSString *)command;

- (redisReply *)execute:(NSString *)command withParam:(id)param;

- (redisReply *)execute:(NSString *)command withKey:(NSString *)key value:(id)value;

- (redisReply *)execute:(NSString *)command withKey:(NSString *)key values:(id)firstValue, ... __attribute__((sentinel));

- (void)setLastErrorID:(NSInteger)id message:(NSString *)msg;

- (id)convertResult:(redisReply *)result;

- (struct timeval)convert_tm:(NSTimeInterval)interval;

@end

@implementation APGRedisServer

#pragma mark Deinitialization

- (void)cleanup {

    if (_context != NULL) {

        redisFree(_context);
        _context = NULL;
    }
}

- (void)dealloc {

    [self cleanup];
}

- (void)disconnect {

    [self cleanup];
}

#pragma mark Initialization

+ (id)connectTo:(NSString *)unixPath {

    APGRedisServer *server = [[APGRedisServer alloc] initWithUnixPath:unixPath];
    return server;
}

+ (id)connectTo:(NSString *)host onPort:(NSUInteger)port {

    APGRedisServer *server = [[APGRedisServer alloc] initWithHost:host onPort:port];
    return server;
}

+ (id)connectTo:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval {

    APGRedisServer *server = [[APGRedisServer alloc] initWithHost:host onPort:port withTimeout:interval];
    return server;
}

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port {

    if (self = [super init]) {

        _context = redisConnect([host cStringUsingEncoding:NSUTF8StringEncoding], (int) port);
    }

    if ((self.context != NULL) && (self.context->err != 0)) {

        [self setLastErrorID:_context->err message:[NSString stringWithCString:_context->errstr encoding:NSUTF8StringEncoding]];
    }

    return self;
}

- (id)initWithHost:(NSString *)host onPort:(NSUInteger)port withTimeout:(NSTimeInterval)interval {

    if (self = [super init]) {

        struct timeval tm = [self convert_tm:interval];

        _context = redisConnectWithTimeout([host cStringUsingEncoding:NSUTF8StringEncoding], (int) port, tm);
    }

    if ((self.context != NULL) && (self.context->err != 0)) {

        [self setLastErrorID:_context->err message:[NSString stringWithCString:_context->errstr encoding:NSUTF8StringEncoding]];
    }

    return self;
}

- (id)initWithUnixPath:(NSString *)unixPath {

    if (self = [super init]) {

        _context = redisConnectUnix([unixPath cStringUsingEncoding:NSUTF8StringEncoding]);
    }

    if ((self.context != NULL) && (self.context->err != 0)) {

        [self setLastErrorID:_context->err message:[NSString stringWithCString:_context->errstr encoding:NSUTF8StringEncoding]];
    }

    return self;
}

#pragma mark Properties

@synthesize lastError = _lastError;

- (BOOL)isConnected {

    if (_context == NULL)
        return FALSE;

    return (_context->err == 0);
}

- (BOOL)requiresAuthentication {

    if (_context == NULL)
        return FALSE;

    redisReply *reply = [self execute:@"CONFIG GET" withParam:@"requirepass"];

    return (reply->len != 0);
}

- (void)setLastErrorID:(NSInteger)id message:(NSString *)msg {

    NSDictionary *d = [NSDictionary dictionaryWithObject:msg forKey:NSLocalizedFailureReasonErrorKey];

    _lastError = [NSError errorWithDomain:NSOSStatusErrorDomain
                                     code:id
                                 userInfo:d];
}

#pragma mark Methods

- (NSString *)ping {

    redisReply *reply = [self execute:@"PING"];

    return [self convertResult:reply];
}

- (NSString *)echo:(NSString *)value {

    redisReply *reply = [self execute:@"ECHO" withParam:value];

    return [self convertResult:reply];
}

- (NSNumber *)dbSize {

    redisReply *reply = [self execute:@"DBSIZE"];

    return [self convertResult:reply];
}

- (BOOL)setDatabase:(NSUInteger)db {

    redisReply *reply = [self execute:@"SELECT" withParam:@(db)];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (BOOL)authenticate:(NSString *)pwd {

    redisReply *reply = [self execute:@"AUTH" withParam:pwd];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (BOOL)hasKey:(NSString *)key {

    redisReply *reply = [self execute:@"EXISTS" withParam:key];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue]);
}

- (BOOL)deleteKey:(NSString *)key {

    redisReply *reply = [self execute:@"DEL" withParam:key];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue]);
}

- (BOOL)increaseTTL:(NSTimeInterval)timeInterval forKey:(NSString *)key {

    redisReply *reply = [self execute:@"EXPIRE" withKey:key value:@(timeInterval)];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (BOOL)persistKey:(NSString *)key {

    redisReply *reply = [self execute:@"PERSIST" withParam:key];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue]);
}

- (NSTimeInterval)retrieveTTL:(NSString *)key {

    redisReply *reply = [self execute:@"TTL" withParam:key];

    NSNumber *result = [self convertResult:reply];

    return [result doubleValue];
}

- (BOOL)renameKey:(NSString *)key toKey:(NSString *)newKey allowOverriding:(BOOL)bOverridingAllowed {

    if (bOverridingAllowed) {

        redisReply *reply = [self execute:@"RENAME" withKey:key value:newKey];

        return [[self convertResult:reply] isEqualToString:@"OK"];
    }

    redisReply *reply = [self execute:@"RENAMENX" withKey:key value:newKey];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue]);
}

- (NSString *)randomKey {

    redisReply *reply = [self execute:@"RANDOMKEY"];

    return [self convertResult:reply];
}

- (BOOL)set:(NSString *)value forKey:(NSString *)key {

    redisReply *reply = [self execute:@"SET" withKey:key value:value];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (BOOL)set:(NSString *)value forKey:(NSString *)key timeOut:(NSTimeInterval)timeOut {

    redisReply *reply = [self execute:@"SETEX"
                              withKey:key
                               values:@(timeOut), value, nil];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (NSString *)get:(NSString *)key {

    redisReply *reply = [self execute:@"GET" withParam:key];

    return [self convertResult:reply];
}

- (BOOL)append:(NSString *)value toKey:(NSString *)key {

    redisReply *reply = [self execute:@"APPEND" withKey:key value:value];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue] > 0);
}

- (NSString *)get:(NSString *)key forRange:(NSRange)range {

    redisReply *reply = [self execute:@"GETRANGE"
                              withKey:key
                               values:@(range.location), @(range.length), nil];

    return [self convertResult:reply];
}

- (BOOL)set:(NSString *)value forKey:(NSString *)key atIndex:(NSUInteger)pos {

    redisReply *reply = [self execute:@"SETRANGE"
                              withKey:key
                               values:@(pos), value, nil];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue] > 0);
}

- (NSUInteger)getLength:(NSString *)key {

    redisReply *reply = [self execute:@"STRLEN" withParam:key];

    NSNumber *result = [self convertResult:reply];

    return [result unsignedIntegerValue];
}

- (BOOL)addCounter:(NSNumber *)value forKey:(NSString *)key {

    redisReply *reply = [self execute:@"SET" withKey:key value:value];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (BOOL)addCounter:(NSNumber *)value forKey:(NSString *)key timeOut:(NSTimeInterval)timeOut {

    redisReply *reply = [self execute:@"SETEX"
                              withKey:key
                               values:@(timeOut), value, nil];

    return [[self convertResult:reply] isEqualToString:@"OK"];
}

- (NSNumber *)getCounter:(NSString *)key {

    redisReply *reply = [self execute:@"GET" withParam:key];

    return [self convertResult:reply];
}

- (NSNumber *)increaseCounter:(NSString *)key {

    return [self increaseCounterBy:@1 forKey:key];
}

- (NSNumber *)increaseCounterBy:(NSNumber *)value forKey:(NSString *)key {

    redisReply *reply = [self execute:@"INCRBY" withKey:key value:value];

    return [self convertResult:reply];
}

- (NSNumber *)decreaseCounter:(NSString *)key {

    return [self decreaseCounterBy:@1 forKey:key];
}

- (NSNumber *)decreaseCounterBy:(NSNumber *)value forKey:(NSString *)key {

    redisReply *reply = [self execute:@"DECRBY" withKey:key value:value];

    return [self convertResult:reply];
}

- (BOOL)createStack:(NSString *)key withItems:(NSArray *)items {

    BOOL result = FALSE;

    for (id elem in items) {

        result = [self pushItem:elem toStack:key];

        if (!result)
            break;
    }

    return result;
}

- (BOOL)createStack:(NSString *)key withItems:(NSArray *)items timeOut:(NSTimeInterval)timeOut {

    if ([self createStack:key withItems:items]) {

        return [self increaseTTL:timeOut forKey:key];
    }

    return FALSE;
}

- (id)popItemFromStack:(NSString *)key {

    redisReply *reply = [self execute:@"RPOP" withParam:key];

    return [self convertResult:reply];
}

- (BOOL)pushItem:(id)item toStack:(NSString *)key {

    redisReply *reply = [self execute:@"LPUSH" withKey:key value:item];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue] > 0);
}

- (BOOL)createQueue:(NSString *)key withItems:(NSArray *)items {

    BOOL result = FALSE;

    for (id elem in items) {

        result = [self pushItem:elem toQueue:key];

        if (!result)
            break;
    }

    return result;
}

- (BOOL)createQueue:(NSString *)key withItems:(NSArray *)items timeOut:(NSTimeInterval)timeOut {

    if ([self createQueue:key withItems:items]) {

        return [self increaseTTL:timeOut forKey:key];
    }

    return FALSE;
}

- (id)popItemFromQueue:(NSString *)key {

    redisReply *reply = [self execute:@"LPOP" withParam:key];

    return [self convertResult:reply];
}

- (BOOL)pushItem:(id)item toQueue:(NSString *)key {

    redisReply *reply = [self execute:@"RPUSH" withKey:key value:item];

    NSNumber *result = [self convertResult:reply];

    return (result && [result intValue] > 0);
}

#pragma mark Help operations

- (id)convertResult:(redisReply *)reply {

    id result = nil;

    if (reply->type == REDIS_REPLY_ERROR) {

        [self setLastErrorID:1000 message:[NSString stringWithCString:reply->str encoding:NSUTF8StringEncoding]];

        freeReplyObject(reply);
        return result;
    }
    else if ((reply->type == REDIS_REPLY_STRING) ||
            (reply->type == REDIS_REPLY_STATUS)) {

        result = [NSString stringWithCString:reply->str encoding:NSUTF8StringEncoding];
        freeReplyObject(reply);
        return result;
    }
    else if (reply->type == REDIS_REPLY_INTEGER) {

        result = [NSNumber numberWithLongLong:reply->integer];
        freeReplyObject(reply);
        return result;
    }
    else if (reply->type == REDIS_REPLY_ARRAY) {

        result = [NSMutableArray arrayWithCapacity:reply->elements];

        for (int i = 0; i < reply->elements; ++i) {

            redisReply *elem = reply->element[i];

            [result setObject:[self convertResult:elem] atIndex:i];
        }

        freeReplyObject(reply);
        return result;
    }
    else {

        freeReplyObject(reply);
        return result;
    }
}

- (redisReply *)execute:(NSString *)command {

    return [self execute:command withParam:nil];
}

- (redisReply *)execute:(NSString *)command withParam:(id)param {

    if (!self.isConnected) return NULL;

    redisReply *reply = NULL;

    if (param) {

        reply = redisCommand(_context, "%s %s", [command cStringUsingEncoding:NSUTF8StringEncoding], [[param description] cStringUsingEncoding:NSUTF8StringEncoding]);
    }
    else {

        reply = redisCommand(_context, "%s", [command cStringUsingEncoding:NSUTF8StringEncoding]);
    }

    return reply;
}

- (redisReply *)execute:(NSString *)command withKey:(NSString *)key value:(id)value {

    if (!self.isConnected) return NULL;

    if ([[value class] isSubclassOfClass:[NSString class]]) {

        return redisCommand(_context, "%s %s %s", [command cStringUsingEncoding:NSUTF8StringEncoding],
                [key cStringUsingEncoding:NSUTF8StringEncoding],
                [value cStringUsingEncoding:NSUTF8StringEncoding]);
    }
    else {

        //TODO: figure out what kind of number is stored [value objCType]
        return redisCommand(_context, "%s %s %ld", [command cStringUsingEncoding:NSUTF8StringEncoding],
                [key cStringUsingEncoding:NSUTF8StringEncoding],
                [value integerValue]);
    }
}

- (redisReply *)execute:(NSString *)command withKey:(NSString *)key values:(id)firstValue, ... {

    if (!self.isConnected) return NULL;

    va_list args;
    id value = firstValue;

    NSMutableString *execCmd = [NSMutableString stringWithFormat:@"%@ %@", command, key];

    va_start (args, firstValue);

    while (value) {

        if ([[value class] isSubclassOfClass:[NSString class]]) {

            [execCmd appendFormat:@" %s", [value cStringUsingEncoding:NSUTF8StringEncoding]];
        }
        else {

            [execCmd appendFormat:@" %@", value];
        }

        value = va_arg (args, id);
    }

    va_end (args);

    return redisCommand(_context, [execCmd cStringUsingEncoding:NSUTF8StringEncoding]);
}

- (struct timeval)convert_tm:(NSTimeInterval)interval {

    struct timeval tm;

    tm.tv_sec = (int) interval;
    tm.tv_usec = (int) ((interval - (int) interval) * 1000000);

    return tm;
}

@end
				