//
//  JSONSerializer.m
//  JSONSerializer
//
//  Created by Andreas Gerlach on 9/22/12.
//  Copyright (c) 2012 Appelgriebsch. All rights reserved.
//

#import "JSONSerializer/JSONSerializer.h"
#import "ReflectionHelper/ReflectionHelper.h"

@implementation JSONSerializer

#pragma mark Public Interface

/*!
    encode the given object as JSON string

    @param theObject - an object that should be encoded

    @return the objects properties and values as JSON encoded string
 */
+ (NSString *)encodeAsString:(id)theObject {

    return [self encodeAsString:theObject error:nil];
}

/*!
    encode the given object as JSON string and providing enhanced error information

    @param theObject - an object that should be encoded

    @param pError    - output parameter that will hold the error information - if any

    @return the objects properties and values as JSON encoded string
 */
+ (NSString *)encodeAsString:(id)theObject error:(NSError **)pError {
    
    NSData* data = [self encodeAsData:theObject error:pError];
    
    if (data) {
        
        return [[NSString alloc] initWithBytes:[data bytes] length:[data length] encoding:NSUTF8StringEncoding];
    }
    
    return nil;
}

/*!
    encode the given object as JSON data

    @param theObject - an object that should be encoded

    @return the objects properties and values as JSON encoded data
 */
+ (NSData *)encodeAsData:(id)theObject {
    
    return [self encodeAsData:theObject error:nil];
}

/*!
    encode the given object as JSON data and providing enhanced error information

    @param theObject - an object that should be encoded

    @param pError    - output parameter that will hold the error information - if any

    @return the objects properties and values as JSON encoded data
 */
+ (NSData *)encodeAsData:(id)theObject error:(NSError **)pError {
    
    NSDictionary *propDict = [ReflectionHelper inspectObject:theObject];
    
    if (propDict) {
    
        [self resolveExternalReferences:propDict];
        
        return [NSJSONSerialization dataWithJSONObject:propDict options:NSJSONWritingPrettyPrinted error:pError];
    }
    
    return nil;    
}

/*!
    decodes a JSON formatted string and returns the object (or dictionary of values)

    @param jsonData - the JSON formatted string to restore the data from

    @return an object that is encoded in the JSON data or a dictionary of key-values
 */
+ (id)decodeString:(NSString *)jsonData {

    return [self decodeString:jsonData error:nil];
}

/*!
    decodes a JSON formatted string and returns the object (or dictionary of values)

    @param jsonData - the JSON formatted string to restore the data from

    @param pError   - output parameter that will hold the error information - if any

    @return an object that is encoded in the JSON data or a dictionary of key-values
 */
+ (id)decodeString:(NSString *)jsonData error:(NSError **)pError {

    NSData* data = [jsonData dataUsingEncoding:NSUTF8StringEncoding];
    
    if (data) {
        
        return [self decodeData:data error:pError];
    }
    
    return nil;
}

/*!
    decodes JSON formatted data and returns the object (or dictionary of values)

    @param jsonData - the JSON formatted data to restore

    @return an object that is encoded in the JSON data or a dictionary of key-values
 */
+ (id)decodeData:(NSData *)jsonData {
    
    return [self decodeData:jsonData error:nil];
}

/*!
    decodes JSON formatted data and returns the object (or dictionary of values)

    @param jsonData - the JSON formatted data to restore

    @param pError   - output parameter that will hold the error information - if any

    @return an object that is encoded in the JSON data or a dictionary of key-values
 */
+ (id)decodeData:(NSData *)jsonData error:(NSError **)pError {
    
    id result = nil;
    NSDictionary *dict = [NSJSONSerialization JSONObjectWithData:jsonData options:0 error:pError];
    
    if (dict) {
        
        result = [ReflectionHelper initObjectWithValues:dict];
    }
    
    return result;
}

#pragma mark Helper Functions

+(void)resolveExternalReferences:(NSDictionary *)dict {
    
    for (NSString* key in [dict allKeys]) {
        
        id value = [dict valueForKey:key];
        
        if (([[value class] isSubclassOfClass:[NSString class]]) ||
            ([[value class] isSubclassOfClass:[NSNumber class]])) {
            
            continue;
        }
        else if ([[value class] isSubclassOfClass:[NSArray class]]) {
            
            NSMutableArray *arr = [NSMutableArray arrayWithCapacity:[value count]];
            
            for (id subElem in value) {
                
                if (([[subElem class] isSubclassOfClass:[NSString class]]) ||
                    ([[subElem class] isSubclassOfClass:[NSNumber class]])) {
                    
                    [arr addObject:subElem];
                }
                else {
                    
                    [arr addObject:[ReflectionHelper inspectObject:subElem]];
                }
            }
            
            [dict setValue:arr forKey:key];
        }
        else if ([[value class] isSubclassOfClass:[NSDictionary class]]) {
            
            [self resolveExternalReferences:value];
        }
        else if ([[value class] isSubclassOfClass:[NSDate class]]) {
            
            [dict setValue:[value description] forKey:key];
        }
        else {
            
            [dict setValue:[ReflectionHelper inspectObject:value] forKey:key];
        }
    }
}

@end
