//
//  ReflectionHelper.m
//  ReflectionHelper
//
//  Created by Andreas Gerlach on 9/22/12.
//  Copyright (c) 2012 Appelgriebsch. All rights reserved.
//

#import <ReflectionHelper/ReflectionHelper.h>
#import "NSObject+Reflection.h"
#import "NSData+Base64.h"

@implementation ReflectionHelper

#pragma mark Public Interface

/*!
    inspects an object and returns a dictionary
    with the property names and values

    @param theObject - the object that should be inspected

    @return a dictionary with all available properties and their values
 */
+ (NSDictionary *)inspectObject:(id)theObject {

    return [theObject properties];
}

/*!
    tries to recreate the object based on the value from the dictionary

    if no suitable class has been found it will return the original dictionary instead

    @param dictValue - the dictionary with the property values from the object

    @return the recreated object (is class is found) or the dictionary itself (if recreation not possible)
 */
+ (id)initObjectWithValues:(NSDictionary *)dictValues {

    if (dictValues) {

        NSString *className = [dictValues valueForKey:DICT_KEY_NAME];

        if (className) {

            return [self initObject:NSClassFromString(className) withValues:dictValues];
        }
    }

    return dictValues;
}

/*!
    tries to recreate the object of the given class based on the value from the dictionary

    @param ofClass   - the class that should be used for re-generating from

    @param dictValue - the dictionary with the property values from the object

    @return the recreated object (is class is found) or an empty instance of the class (if recreation not possible)

 */
+ (id)initObject:(Class)ofClass withValues:(NSDictionary *)dictValues {

    id result = nil;

    if ((ofClass) &&
            (dictValues)) {

        result = [[ofClass alloc] init];

        for (NSString *key in dictValues) {

            id value = [dictValues objectForKey:key];

            if (![key hasPrefix:@"__"]) {

                // check for sub objects
                if ([[value class] isSubclassOfClass:[NSDictionary class]]) {

                    NSString *subClass = [value objectForKey:DICT_KEY_NAME];

                    if (subClass != nil) {

                        value = [self initObject:NSClassFromString(subClass) withValues:value];
                    }
                }
                else if ([[value class] isSubclassOfClass:[NSArray class]]) {

                    NSMutableArray *subValues = [NSMutableArray arrayWithCapacity:[value count]];

                    for (id subElem in value) {

                        // check for sub objects
                        if ([[subElem class] isSubclassOfClass:[NSDictionary class]]) {

                            NSString *subClass = [subElem objectForKey:DICT_KEY_NAME];

                            if (subClass != nil) {

                                [subValues addObject:[self initObject:NSClassFromString(subClass) withValues:subElem]];
                            }
                        }
                        else {

                            [subValues addObject:subElem];
                        }
                    }

                    value = subValues;
                }

                [result setValue:value forKey:key];
            }
            else if ([key isEqualToString:DICT_KEY_DATA]) {

                NSData *data = [NSData dataWithBase64EncodedString:value];
                result = [NSKeyedUnarchiver unarchiveObjectWithData:data];
            }
        }
    }

    return result;
}

@end
